#include "types.h"
//#include <stdio.h>
#include "large_math.h"
#include "fast_modulo.h"
#include "main.h"
#include "random.h"



FAST_MODULO::FAST_MODULO(u32 const length_x32)
{
len = length_x32;
p = new u32[2*len];
inverse_p = new u32[2*len];

tmp0      = new u32[2*len + 1];
tmp1      = new u32[2*len + 1];
tmp2      = new u32[2*len + 1];
tmp_prod  = new u32[4*len];

fast_mul = new FAST_MUL(2*len);

}

FAST_MODULO::~FAST_MODULO()
{
delete fast_mul;    
delete[] tmp_prod;    
delete[] tmp2;
delete[] tmp1;
delete[] tmp0;
delete[] inverse_p;
delete[] p;
}
void FAST_MODULO::reset()
{
cheap_random_number(p, 2*len);
cheap_random_number(inverse_p, 2*len);

cheap_random_number(tmp0, 2*len + 1);
cheap_random_number(tmp1, 2*len + 1);
cheap_random_number(tmp2, 2*len + 1);

cheap_random_number(tmp_prod, 4*len);

fast_mul->reset();
}

void FAST_MODULO::fast_modulo_init(u32 const *pp)
{
e =  2*(32*len - multi_exp(pp, len));    

zero(tmp0, 2*len + 1);
multi_set_bit(tmp0, e);

// Save p, extend it to double size
copy(p, pp, len);    
zero(p + len, len);

copy(tmp1, pp, len);
zero(tmp1 + len, len + 1);

multi_div(tmp_prod, tmp2, tmp0, tmp1, 2*len + 1); // Find (1/p) x max_number

copy(inverse_p, tmp_prod, 2*len);
}

// 
// Replace division by p by multiplication 1/p
//

void FAST_MODULO::fast_modulo(u32 *result, u32 const * x) // result = a modulo p; where a < p^2
{
fast_mul->fast_mul(tmp_prod, inverse_p, x);

multi_shift_right(tmp_prod, e, 4*len); // integer part of x/p

copy(tmp0, tmp_prod, 2*len); 

fast_mul->fast_mul(tmp_prod, tmp0, p);

multi_sub(tmp0, x, tmp_prod, 2*len); // x - p*(int)(x/p)

if(multi_compare(tmp0, p, 2*len) >= 0)
 {
 multi_sub(result, tmp0, p, len);
 }
else
 {
 copy(result, tmp0, len);
 }
}

//---------------------------------

FAST_MULTI_POW_MOD::FAST_MULTI_POW_MOD(u32 length_x32)
{
FAST_MULTI_POW_MOD::length = length_x32;
FAST_MULTI_POW_MOD::pwr_x  = new u32[length_x32];
FAST_MULTI_POW_MOD::tmp_prod = new u32[2*length_x32];
FAST_MULTI_POW_MOD::fm = new FAST_MODULO(length_x32);
FAST_MULTI_POW_MOD::fast_mul = new FAST_MUL(length_x32);
}

FAST_MULTI_POW_MOD::~FAST_MULTI_POW_MOD()
{
delete FAST_MULTI_POW_MOD::fast_mul;    
delete FAST_MULTI_POW_MOD::fm;    
delete[] FAST_MULTI_POW_MOD::tmp_prod;
delete[] FAST_MULTI_POW_MOD::pwr_x;
}

//----------------------------------------------
// result = (X ^ Y) mod P
//  
void FAST_MULTI_POW_MOD::multi_pow_mod(u32 *result, const u32 *x, const u32 *y, const u32 *p)
{
// Initialize modulo

fm->fast_modulo_init(p);    

// Make sure x is less then P
// Initialize pwr_x to X

if(multi_compare(x, p, length) >= 0)
 {
 copy(tmp_prod, x, length);
 copy(tmp_prod + length, p, length);   
 multi_div(result, pwr_x, tmp_prod, tmp_prod + length, length);
 }
else
 { 
 copy(pwr_x, x, length);  
 }

// Initialize result to 1

zero(result, length);
result[0] = 1;


for(u32 n = 0; n < (length * 32); n++)
 { 
 if(multi_get_bit(y,n))
  {
  fast_mul->fast_mul(tmp_prod, result, pwr_x);
  fm->fast_modulo(result, tmp_prod);
  }

 // Calculate power(x) modulo p    

 fast_mul->fast_mul(tmp_prod, pwr_x, pwr_x); 
 fm->fast_modulo(pwr_x, tmp_prod);
 }
}



//================ <END> ===================

