#include <tchar.h>
#include "control.h"
#include "resource.h"
#include "main.h"
//----------------------------------------------------------------
// 0 is reserved for connection request
static const u32 CMD_START_CONNECTION = 1;
static const u32 CMD_CONNECTED = 2;
static const u32 CMD_SHOW_MESSAGE = 3;
static const u32 CMD_MESSAGE_RECEIVED = 4;
static const u32 CMD_HANG_UP = 5;
static const u32 CMD_HANG_DOWN = 6;
static const u32 CMD_IN_VOICE = 7;
static const u32 CMD_HOLD = 8;
static const u32 CMD_VOICE = 9;
//----------------------------------------------------------------
CONTR_PROT::CONTR_PROT(NODE* node, HWND mWnd, HWND parent)
{
	mainWnd = mWnd;
	this->parent = parent;
    nd = *node;
    state = CTRL_ST_IDLE;
    msg = NULL;
    send_msg_num = 0;
    rec_msg_num = 0;
	rec_msg_flag = 0;
    for (int i = 0; i < CONTR_MAX_VOICE_BUF; i++)
	{
        rec_data[i] = NULL;
        rec_size[i] = 0;
	}
	send_voice_buf = NULL;
	voice_cnt = 0;
}
//----------------------------------------------------------------
CONTR_PROT::~CONTR_PROT()
{
	if (send_voice_buf) delete[] send_voice_buf;
    if (msg)
    {
        MESSAGE* cur;
        for (cur=msg;cur;)
        {
            MESSAGE* next = cur->next;
			delete cur->text;
            delete cur;
            cur = next;
        }
    }
}
//----------------------------------------------------------------
static void TransferMessage(HWND hWnd, MESSAGE* message,NODE* nd)
{
	int len = _tcslen(message->text)*sizeof(TCHAR);
	u32 buf[NETWORK_SIZE_USEFUL/4] = {0};
    buf[0] = CMD_SHOW_MESSAGE;
    buf[1] = message->msg_num;
	if (len > NETWORK_SIZE_USEFUL-8) len = NETWORK_SIZE_USEFUL-8;
	CopyMemory(&buf[2],message->text,len);
	PACKET_FOR_SEND pack;
	pack.buffer = buf;
	pack.buf_len = 2*sizeof(u32) + len;
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	getNet()->send_packet(&pack);
    message->send_cnt++;
}
//----------------------------------------------------------------
void CONTR_PROT::sendMessage(TCHAR* message)
{
    MESSAGE* ptr;
    if (state == CTRL_ST_IDLE) {
		state = CTRL_ST_CONNECTION;
		process();
        send_msg_num = 0;
        rec_msg_num = 0;
	}
    if (msg)
    {
        for (ptr = msg; ptr->next; ptr = ptr->next);
        ptr->next = new MESSAGE;
        ptr = ptr->next;
    }
    else
    {
        msg = new MESSAGE;
        ptr = msg;
    }
    ptr->msg_num = send_msg_num++;
    ptr->next = NULL;
    ptr->send_cnt = 0;
    ptr->text = new TCHAR[_tcslen(message)+1];
	_tcscpy_s(ptr->text,_tcslen(message)+1,message);
    TransferMessage(mainWnd,ptr,&nd);
}
//----------------------------------------------------------------
void CONTR_PROT::sendVoice(u8* data, u32 size)
{
	PACKET_FOR_SEND pack;
	if (state != CTRL_ST_IN_VOICE) return;
	if (!send_voice_buf) 
	{
		send_voice_buf = new u32[size+1];
	    send_voice_buf[0] = CMD_VOICE;
	}
	CopyMemory(&send_voice_buf[1],data,size);
	pack.buffer = send_voice_buf;
	pack.buf_len = sizeof(u32) + size;
	pack.nd = &nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	getNet()->send_packet(&pack);
	voice_cnt++;
}
//----------------------------------------------------------------
void CONTR_PROT::processMessages(void)
{
    MESSAGE* ptr;
    for (ptr = msg; ptr; ptr = ptr->next)
    {
        TransferMessage(parent,ptr,&nd);
    }
}

//----------------------------------------------------------------


void CONTR_PROT::process(void)
{
	PACKET_FOR_SEND pack;
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;
	NETWORK* net = getNet();

    switch(state)
    {
    case CTRL_ST_IDLE:
    default:
        break;
    case CTRL_ST_CONNECTION:
		pack.buffer = &CMD_START_CONNECTION;
		pack.buf_len = sizeof(CMD_START_CONNECTION);
		if (net) net->send_packet(&pack);
        break;
    case CTRL_ST_CONNECTED:
    case CTRL_ST_DIALING:
    case CTRL_ST_IN_VOICE:
    case CTRL_ST_HOLD:
        processMessages();
        break;
    }
    if (state == CTRL_ST_DIALING)
	{
		pack.buffer = &CMD_HANG_UP;
		pack.buf_len = sizeof(CMD_HANG_UP);
		if (net) net->send_packet(&pack);
	}
    if (state == CTRL_ST_IN_VOICE)
    {
        voice_cnt++;
		pack.buffer = &CMD_IN_VOICE;
		pack.buf_len = sizeof(CMD_IN_VOICE);
		if (net) net->send_packet(&pack);
        if (voice_cnt > 20)
        {
            state = CTRL_ST_DIALING;
        }
    }
    if (state == CTRL_ST_HOLD)
    {
        voice_cnt++;
		pack.buffer = &CMD_HOLD;
		pack.buf_len = sizeof(CMD_HOLD);
		if (net) net->send_packet(&pack);
        if (voice_cnt > 20)
        {
            state = CTRL_ST_DIALING;
        }
    }
}

void CONTR_PROT::received(u32* packet, u32 size)
{
 	PACKET_FOR_SEND pack;
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;
	NETWORK* net = getNet();

   if (size < sizeof(u32)) return;
    switch(packet[0]) // command
    {
    case CMD_START_CONNECTION:
		pack.buffer = &CMD_CONNECTED;
		pack.buf_len = sizeof(CMD_CONNECTED);
		if (net) net->send_packet(&pack);
        if (state != CTRL_ST_CONNECTED)
        {
            state = CTRL_ST_CONNECTED;
            send_msg_num = 0;
            rec_msg_num = 0;
        }
        break;
    case CMD_CONNECTED:
        if (state < CTRL_ST_CONNECTED)
        {
            state = CTRL_ST_CONNECTED;
        }
        break;
    case CMD_SHOW_MESSAGE:
        if(size > sizeof(u32)*2)
        {
			// send responce
            u32 data[2] = {CMD_MESSAGE_RECEIVED,packet[1]};
			pack.buffer = data;
			pack.buf_len = sizeof(data);
			if (net) net->send_packet(&pack);
			// parse packet
			u32 cur_num = packet[1];
			if (cur_num >= rec_msg_num)
            {
				SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[2],size-sizeof(u32)*2,0,100,NULL);
				rec_msg_flag >>= cur_num - rec_msg_num;
                rec_msg_num = cur_num;
				rec_msg_flag |= 0x80000000;
            }
			else if (!rec_msg_num || !cur_num || ((rec_msg_num - cur_num) > 32)) // session restart
			{
				SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[2],size-sizeof(u32)*2,0,100,NULL);
				rec_msg_num = cur_num;
				rec_msg_flag = 0x80000000;
			}
			else
			{
				u8 shift = rec_msg_num - cur_num;
				u32 mask = 0x80000000 >> shift;
				if (!(rec_msg_flag & mask))
				{
					SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[2],size-sizeof(u32)*2,0,100,NULL);
				}
				rec_msg_flag |= mask;
			}
        }
        break;
    case CMD_MESSAGE_RECEIVED:
        if(size >= sizeof(u32)*2)
        {
            MESSAGE *ptr, *prev_ptr;
            for (prev_ptr = ptr = msg; ptr; ptr = ptr->next)
            {
                if (ptr->msg_num == packet[1])
                {
                    if (prev_ptr == ptr) // very first item
                    {
                        msg = ptr->next;
                    }
                    else
                    {
                        prev_ptr->next = ptr->next;
                    }
					delete ptr->text;
                    delete ptr;
                    break;
                }
            }
        }
        break;
    case CMD_HANG_UP:
        state = CTRL_ST_IN_VOICE;
        voice_cnt = 0;
		pack.buffer = &CMD_IN_VOICE;
		pack.buf_len = sizeof(CMD_IN_VOICE);
		if (net) net->send_packet(&pack);
        break;
    case CMD_HANG_DOWN:
        state = CTRL_ST_CONNECTED;
        break;
    case CMD_IN_VOICE:
        voice_cnt = 0;
        if (state == CTRL_ST_DIALING)
            state = CTRL_ST_IN_VOICE;
        else if (state < CTRL_ST_IN_VOICE)
		{
			pack.buffer = &CMD_HANG_DOWN;
			pack.buf_len = sizeof(CMD_HANG_DOWN);
			if (net) net->send_packet(&pack);
		}
        break;
    case CMD_HOLD:
        if (state == CTRL_ST_IN_VOICE)
            state = CTRL_ST_HOLD;
        else if (state != CTRL_ST_HOLD)
		{
			pack.buffer = &CMD_HANG_DOWN;
			pack.buf_len = sizeof(CMD_HANG_DOWN);
			if (net) net->send_packet(&pack);
		}
        break;
    case CMD_VOICE:
        if (state == CTRL_ST_IN_VOICE)
        {
            // TODO: critical section is required
            int inx;
            for (inx = 0; inx < CONTR_MAX_VOICE_BUF; inx++)
            {
                if (!rec_data[inx]) break;
            }
            if (inx < CONTR_MAX_VOICE_BUF)
            {
                rec_size[inx] = size-sizeof(u32);
                rec_data[inx] = new u8[rec_size[inx]];
                memcpy(rec_data[inx],&packet[1],rec_size[inx]);
            }
        }
        break;
	}
}


void CONTR_PROT::dial(void)
{
    if (state < CTRL_ST_DIALING) state = CTRL_ST_DIALING;
    else state = CTRL_ST_CONNECTED;
}

u8* CONTR_PROT::getVoice(u32* size)
{
	*size = 0;
    if ((state != CTRL_ST_IN_VOICE) || !rec_data[0]) return NULL;

    // TODO: critical section
    u8* buf = rec_data[0];
    *size = rec_size[0];
    memmove(rec_data,&rec_data[1],(CONTR_MAX_VOICE_BUF-1)*sizeof(u8*));
    memmove(rec_size,&rec_size[1],(CONTR_MAX_VOICE_BUF-1)*sizeof(u32));
    rec_data[CONTR_MAX_VOICE_BUF-1] = NULL;
    rec_size[CONTR_MAX_VOICE_BUF-1] = 0;
	return buf;
}
