#include <windows.h>
#include <commctrl.h>
#include <tchar.h>
#include <Windowsx.h>

#include "resource.h"
#include "settings.h"
#include "password.h"
#include "network.h"
#include "kademlia.h"
#include "chat.h"
#include "build.h"
#include "upnp.h"
#include "log.h"
#include "users.h"
#include "lists.h"
#include "main.h"
#include "file_transfer.h"
#include "..\audio\waveio.h"
//----------------------------------------------------------------------------
static LPTSTR module = TEXT("Main");
//----------------------------------------------------------------------------
typedef struct {
	HWND hWnd;
	u32 id[NETWORK_ID_LENGTH];
} chatwnd_t;
//----------------------------------------------------------------------------
typedef struct {
	CRITICAL_SECTION cs;
	u8 password[password_max_length];
	u8 passwordMix[password_max_length];
	u32 passwordLength;
	u32 passwordTimeOut;
	HANDLE thPassword;
	HANDLE eStopThreadPassword;
	SECRET_DATA sdata;
	KADEMLIA* kad;
	NETWORK* net;
	USERS* usr;
	FILE_TRANSFER* ft;
	CRITICAL_SECTION chat_cs;
	chatwnd_t* chats;
	u32 chats_num;
	HANDLE thVoice;
	HANDLE eStop;
	HWND hMainWnd;
	bool restart;
	UINT wiNum;
	UINT woNum;
} instance_t;
//----------------------------------------------------------------------------
static WAVEIN* win;
static WAVEOUT* wout;
static bool bWindows7;
static instance_t* inst; // must be static for access without message transfer
//----------------------------------------------------------------------------
KADEMLIA* getKad(void)
{
	return inst->kad;
}
//----------------------------------------------------------------------------
NETWORK* getNet(void)
{
	return inst->net;
}
//----------------------------------------------------------------------------
SECRET_DATA* getSecretData(void)
{
	return &inst->sdata;
}
//----------------------------------------------------------------------------
FILE_TRANSFER* getFileTransfer(void)
{
	return inst->ft;
}
//----------------------------------------------------------------------------
void getFileTransferPath(LPTSTR FileName)
{
	settings_t *set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	cp_strcchcpy(FileName, MAX_PATH, set->ftPath);
	cp_strcchcpy(&FileName[cp_strlen(FileName)], MAX_PATH - cp_strlen(FileName), TEXT("\\"));
}
//----------------------------------------------------------------------------
DWORD WINAPI voiceProc(instance_t* inst)
{
	s16 *buf_in, *buf_out;

	OSVERSIONINFO ver;
	ver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&ver);
	bWindows7 = ver.dwMajorVersion > 5;
	win = new WAVEIN(inst->wiNum);
	wout = new WAVEOUT(inst->woNum);

	while(WAIT_TIMEOUT == WaitForSingleObject(inst->eStop,0))
	{
		if (inst->restart)
		{
			if (win) delete win;
			if (wout) delete wout;
			win = new WAVEIN(inst->wiNum);
			wout = new WAVEOUT(inst->woNum);
			inst->restart = false;
		}
		if (win && wout)
		{
			if (WAIT_TIMEOUT != WaitForSingleObject(win->rxEvent,100) && (buf_in = (s16*)win->GetBuffer()))
			{
				buf_out = (s16*)wout->GetBuffer();
				if (buf_out) memset(buf_out,0,sizeof(s16)*BufferSize);
				//EnterCriticalSection(&inst->chat_cs);
				for (u32 i = 0; i < inst->chats_num; i++)
				{
					if (inst->chats[i].hWnd) SendMessageTimeout(inst->chats[i].hWnd,WM_CHANGE_VOICE,(WPARAM)buf_in,(LPARAM)buf_out,0,100,NULL);
				}
				//LeaveCriticalSection(&inst->chat_cs);
			}
		}
		else
		{
			inst->restart = true;
			Sleep(100);
		}
	}
	if (win) delete win;
	if (wout) delete wout;

	return 0;
}
//----------------------------------------------------------------------------
void setVolume(WORD value)
{
	DWORD volume;
	if (value >= 100) volume = 0xFFFF;
	else volume = (DWORD)value*0xFFFF/100;
	volume |= volume<<16;
	waveOutSetVolume(wout->GetWaveOutHandle(),volume);
}
//----------------------------------------------------------------------------
WORD getVolume(void)
{
	DWORD volume;
	waveOutGetVolume(wout->GetWaveOutHandle(),&volume);
	return (WORD)((volume&0xFFFF)*100/0xFFFF);
}
//----------------------------------------------------------------------------
void SetWaveParam(settings_t* set)
{
	inst->wiNum = WAVE_MAPPER;
	for (UINT i = 0; i < waveInGetNumDevs(); i++)
	{
		WAVEINCAPS cap;
		if (MMSYSERR_NOERROR == waveInGetDevCaps(i,&cap,sizeof(cap)))
		{
			if (!cp_strcmp(set->microphone,cap.szPname))
			{
				inst->wiNum = i;
				break;
			}
		}
	}
	inst->woNum = WAVE_MAPPER;
	for (UINT i = 0; i < waveOutGetNumDevs(); i++)
	{
		WAVEOUTCAPS cap;
		if (MMSYSERR_NOERROR == waveOutGetDevCaps(i,&cap,sizeof(cap)))
		{
			if (!cp_strcmp(set->speaker,cap.szPname))
			{
				inst->woNum = i;
				break;
			}
		}
	}
	inst->restart = true;
}
//----------------------------------------------------------------------------
// Set the record input, volume and balance
static void MixerParametersXP(HWAVEIN hWaveIn, WORD* volume_in_percent, bool set)
{
 UINT i;
 HMIXER mix;
 MIXERLINE mixerDstLine;
 MIXERLINE mixerSrcLine;
 MIXERCONTROL mixerCtrl;
 MIXERLINECONTROLS mixerLineCtrl;
 MIXERCONTROLDETAILS mixerCtrlDetails;
 MIXERCONTROLDETAILS_UNSIGNED mixerCtrlDetailsUnsigned;

 // Open the mixer
 if (MMSYSERR_NOERROR != mixerOpen(&mix, (UINT)hWaveIn, 0, 0, MIXER_OBJECTF_HWAVEIN))
  {
  return;
  }
  
 // Get wavein
 mixerDstLine.cbStruct = sizeof(MIXERLINE);
 mixerDstLine.dwComponentType = MIXERLINE_COMPONENTTYPE_DST_WAVEIN;

 if (MMSYSERR_NOERROR != mixerGetLineInfo((HMIXEROBJ)mix, &mixerDstLine, MIXER_GETLINEINFOF_COMPONENTTYPE))
  {
  mixerClose(mix);
  return;
  }

 // Get line input
 for (i = 0; i < mixerDstLine.cConnections; i++)
 {
  mixerSrcLine.cbStruct = sizeof(MIXERLINE);
  mixerSrcLine.dwDestination = mixerDstLine.dwDestination;
  mixerSrcLine.dwSource = i;

  mixerGetLineInfo((HMIXEROBJ)mix, &mixerSrcLine, MIXER_GETLINEINFOF_SOURCE);

  // if source is line in 
  if (mixerSrcLine.dwComponentType == MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE) break;
 }

 if (mixerSrcLine.dwComponentType != MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE)
  {
  mixerClose(mix);
  return;
  }

 // Get switch control
 mixerLineCtrl.cbStruct = sizeof(MIXERLINECONTROLS);
 mixerLineCtrl.dwLineID = mixerDstLine.dwLineID;
 mixerLineCtrl.dwControlType = MIXERCONTROL_CONTROLTYPE_MUX;
 mixerLineCtrl.cControls = 1;
 mixerLineCtrl.pamxctrl = &mixerCtrl;
 mixerLineCtrl.cbmxctrl = sizeof(MIXERCONTROL);

 if (MMSYSERR_NOERROR != mixerGetLineControls((HMIXEROBJ)mix, &mixerLineCtrl, MIXER_GETLINECONTROLSF_ONEBYTYPE))
  {
  mixerClose(mix);
  return;
  }

 // select line input
 if (mixerCtrl.fdwControl & MIXERCONTROL_CONTROLF_MULTIPLE) // multiple controle case
 {
  MIXERCONTROLDETAILS_UNSIGNED *mux_vals = new MIXERCONTROLDETAILS_UNSIGNED[mixerCtrl.cMultipleItems];

  for (i = 0; i < mixerCtrl.cMultipleItems; i++)
  {
   if (mixerSrcLine.dwSource == i) mux_vals[i].dwValue = 1; // set corresponding input to enable state
   else mux_vals[i].dwValue = 0;
  }

  mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
  mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
  mixerCtrlDetails.cChannels = 1;
  mixerCtrlDetails.cMultipleItems = mixerCtrl.cMultipleItems;
  mixerCtrlDetails.paDetails = mux_vals;
  mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);

  if (MMSYSERR_NOERROR != mixerSetControlDetails((HMIXEROBJ)mix, &mixerCtrlDetails, MIXER_SETCONTROLDETAILSF_VALUE))
   {
   mixerClose(mix);
   delete[] mux_vals;
   return;
   }

  delete[] mux_vals;
 }
 else // single controle case
 {
  mixerCtrlDetailsUnsigned.dwValue = mixerSrcLine.dwSource;

  mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
  mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
  mixerCtrlDetails.cMultipleItems = 0;
  mixerCtrlDetails.paDetails = &mixerCtrlDetailsUnsigned;
  mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
  mixerCtrlDetails.cChannels = 1;

  if (MMSYSERR_NOERROR != mixerSetControlDetails((HMIXEROBJ)mix, &mixerCtrlDetails, MIXER_SETCONTROLDETAILSF_VALUE))
   {
   mixerClose(mix);
   return;
   }
 }

 // Get the volume control
 mixerLineCtrl.cbStruct = sizeof(MIXERLINECONTROLS);
 mixerLineCtrl.dwLineID = mixerSrcLine.dwLineID;
 mixerLineCtrl.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
 mixerLineCtrl.cControls = 1;
 mixerLineCtrl.pamxctrl = &mixerCtrl;
 mixerLineCtrl.cbmxctrl = sizeof(MIXERCONTROL);

 if (MMSYSERR_NOERROR != mixerGetLineControls((HMIXEROBJ)mix, &mixerLineCtrl, MIXER_GETLINECONTROLSF_ONEBYTYPE))
  {
  mixerClose(mix);
  return;
  }

mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
mixerCtrlDetails.cMultipleItems = 0;
mixerCtrlDetails.paDetails = &mixerCtrlDetailsUnsigned;
mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
mixerCtrlDetails.cChannels = 1;
 if (set)
 {
	 // Set volume
	 mixerCtrlDetailsUnsigned.dwValue = mixerCtrl.Bounds.dwMinimum + (mixerCtrl.Bounds.dwMaximum - mixerCtrl.Bounds.dwMinimum)*(*volume_in_percent)/100;

	 if (MMSYSERR_NOERROR != mixerSetControlDetails((HMIXEROBJ)mix, &mixerCtrlDetails, MIXER_SETCONTROLDETAILSF_VALUE))
	  {
	  mixerClose(mix);
	  return;
	  }
 }
 else
 {
	 // Get volume
	 if (MMSYSERR_NOERROR != mixerGetControlDetails((HMIXEROBJ)mix, &mixerCtrlDetails, MIXER_GETCONTROLDETAILSF_VALUE))
	  {
	  mixerClose(mix);
	  return;
	  }
	 *volume_in_percent = (WORD)((mixerCtrlDetailsUnsigned.dwValue - mixerCtrl.Bounds.dwMinimum)*100/(mixerCtrl.Bounds.dwMaximum - mixerCtrl.Bounds.dwMinimum));
 }

 // close the mixer
 mixerClose(mix);
}

// Set the record input, volume and balance
static void MixerParametersW7(HWAVEIN hWaveIn, WORD* volume_in_percent, bool set)
{
 HMIXER mix;
 MIXERLINE mixerLine;
 MIXERCONTROL mixerCtrl;
 MIXERLINECONTROLS mixerLineCtrl;
 MIXERCONTROLDETAILS mixerCtrlDetails;
 MIXERCONTROLDETAILS_UNSIGNED mixerCtrlDetailsUnsigned;

 // Open the mixer
 if (MMSYSERR_NOERROR != mixerOpen(&mix, (UINT)hWaveIn, 0, 0, MIXER_OBJECTF_HWAVEIN))
  {
  return;
  }
  
 // Get wavein
 mixerLine.cbStruct = sizeof(MIXERLINE);
 mixerLine.dwComponentType = MIXERLINE_COMPONENTTYPE_DST_WAVEIN;

 if (MMSYSERR_NOERROR != mixerGetLineInfo((HMIXEROBJ)mix, &mixerLine, MIXER_GETLINEINFOF_COMPONENTTYPE))
  {
  mixerClose(mix);
  return;
  }

 // Get the volume control
 mixerLineCtrl.cbStruct = sizeof(MIXERLINECONTROLS);
 mixerLineCtrl.dwLineID = mixerLine.dwLineID;
 mixerLineCtrl.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
 mixerLineCtrl.cControls = 1;
 mixerLineCtrl.pamxctrl = &mixerCtrl;
 mixerLineCtrl.cbmxctrl = sizeof(MIXERCONTROL);

 if (MMSYSERR_NOERROR != mixerGetLineControls((HMIXEROBJ)mix, &mixerLineCtrl, MIXER_GETLINECONTROLSF_ONEBYTYPE))
  {
  mixerClose(mix);
  return;
  }

mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
mixerCtrlDetails.cMultipleItems = 0;
mixerCtrlDetails.paDetails = &mixerCtrlDetailsUnsigned;
mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
mixerCtrlDetails.cChannels = 1;
 if (set)
 {
	 // Set volume
	 mixerCtrlDetailsUnsigned.dwValue = mixerCtrl.Bounds.dwMinimum + (mixerCtrl.Bounds.dwMaximum - mixerCtrl.Bounds.dwMinimum)*(*volume_in_percent)/100;

	 if (MMSYSERR_NOERROR != mixerSetControlDetails((HMIXEROBJ)mix, &mixerCtrlDetails, MIXER_SETCONTROLDETAILSF_VALUE))
	  {
	  mixerClose(mix);
	  return;
	  }
 }
 else
 {
	 // Get volume
	 if (MMSYSERR_NOERROR != mixerGetControlDetails((HMIXEROBJ)mix, &mixerCtrlDetails, MIXER_GETCONTROLDETAILSF_VALUE))
	  {
	  mixerClose(mix);
	  return;
	  }
	 *volume_in_percent = (WORD)((mixerCtrlDetailsUnsigned.dwValue - mixerCtrl.Bounds.dwMinimum)*100/(mixerCtrl.Bounds.dwMaximum - mixerCtrl.Bounds.dwMinimum));
 }

 // close the mixer
 mixerClose(mix);
}
//----------------------------------------------------------------------------
void setMicrophone(WORD value)
{
  if (bWindows7)
  {
   MixerParametersW7(win->GetWaveInHandle(), &value,true); // set mixer's mux to line and volume (in %)
  }
  else
  {
   MixerParametersXP(win->GetWaveInHandle(), &value,true); // set mixer's mux to line and volume (in %)
  }
}
//----------------------------------------------------------------------------
WORD getMicrophone(void)
{
	WORD value;
  if (bWindows7)
  {
   MixerParametersW7(win->GetWaveInHandle(), &value,false); // set mixer's mux to line and volume (in %)
  }
  else
  {
   MixerParametersXP(win->GetWaveInHandle(), &value,false); // set mixer's mux to line and volume (in %)
  }
  return value;
}

//----------------------------------------------------------------------------
DWORD WINAPI PasswordThread(void* param)
{
	instance_t* inst = (instance_t*)param;
	settings_t *set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	while (WAIT_TIMEOUT == WaitForSingleObject(inst->eStopThreadPassword, 60000))
	{
		EnterCriticalSection(&inst->cs);
		if (!inst->passwordTimeOut && inst->passwordLength && set->passwordTime) // reset password
		{
			ZeroMemory(inst->password,sizeof(inst->password));
			ZeroMemory(inst->passwordMix,sizeof(inst->password));
			inst->passwordLength = 0;
			LeaveCriticalSection(&inst->cs);
			LogMessage(module,TEXT("Password Cleared"));
		}
		else
		{
			u32 i;
			if (inst->passwordTimeOut)
			{
				inst->passwordTimeOut--;
			}
			for (i = 0; i < inst->passwordLength; i++)
			{
				inst->password[i] ^= inst->passwordMix[i];
			}
			cheap_random_number(inst->passwordMix, (inst->passwordLength+3)/4);
			for (i = 0; i < inst->passwordLength; i++)
			{
				inst->password[i] ^= inst->passwordMix[i];
			}
			LeaveCriticalSection(&inst->cs);
			LogMessage(module,TEXT("Password Mixed"));
		}
	}
	return 0;
}
//----------------------------------------------------------------------------
u32 getPassword(u8* password)
{
	u32 pass_len = 0;
	u32 i;
	if (!inst->sdata.data_is_initialized) return 0;
	// try to get remembered password
	EnterCriticalSection(&inst->cs);
	if (inst->passwordLength)
	{
		for (i = 0; i < inst->passwordLength; i++)
		{
			password[i] = inst->password[i]^inst->passwordMix[i];
		}
		pass_len = inst->passwordLength;
	}
	LeaveCriticalSection(&inst->cs);

	if (pass_len) 
	{
		LogMessage(module,TEXT("Get password: stored value"));
		return pass_len; // password exists
	}

	u32 cnt = 5;
	while(cnt--)
	{
		pass_len = PasswordRequest(inst->hMainWnd,password);
		if (!pass_len) return 0;
		if (check_password(&inst->sdata.encoded_secret_key,password,pass_len))
		{
			settings_t *set = (settings_t*)GetProp(inst->hMainWnd,TEXT("set"));
			if (set->passwordTime)
			{
				EnterCriticalSection(&inst->cs);
				inst->passwordTimeOut = set->passwordTime;
				inst->passwordLength = pass_len;
				cheap_random_number(inst->passwordMix, pass_len);
				for (i = 0; i < inst->passwordLength; i++)
				{
					inst->password[i] = password[i]^inst->passwordMix[i];
				}
				LeaveCriticalSection(&inst->cs);
			}
			LogMessage(module,TEXT("Get password: user is requested"));
			return pass_len;
		}
		else MessageBox(inst->hMainWnd,TEXT("Password doesn't match key\nTry again"),TEXT("Password"),MB_OK|MB_ICONEXCLAMATION);
	}
	MessageBox(inst->hMainWnd,TEXT("Too many attempts to enter the password\nPROGRAM WILL BE CLOSED"),TEXT("Error"),MB_OK|MB_ICONERROR);
	LogMessage(module,TEXT("USER ENTER WRONG PASSWORD MANY TIMES PROGRAM WILL BE CLOSED"));
	DestroyWindow(inst->hMainWnd);
	return 0;
}
//----------------------------------------------------------------------------
void parsePacket(NET_PACKET* pac)
{
	u32 i;
    switch(pac->buffer[0])
    {
    case NETWORK_PROT_KADEMLIA:
		inst->kad->packetReceived(&pac->buffer[1],pac->buf_len-sizeof(u32),pac->nd,inst->usr);
        break;
    case NETWORK_PROT_CONTROL:
		if ((pac->buf_len >= 2 * sizeof(u32)) && !pac->buffer[1]) // connection request
		{
			SendMessage(inst->hMainWnd, WM_REQUEST_CONNECTION, 0, (LPARAM)pac->nd);
			break;
		}
		EnterCriticalSection(&inst->chat_cs);
		for (i = 0; i < inst->chats_num; i++)
		{
			if (inst->chats[i].hWnd && isIdIdentical(pac->nd->id,inst->chats[i].id)) break;
		}
		if (i >= inst->chats_num) // create new chat
		{
			LeaveCriticalSection(&inst->chat_cs);
			// have to create new window in appropriate thread!!!
			SendMessage(inst->hMainWnd,WM_CREATECHAT,0,(LPARAM)pac->nd->id);
			break;
			/*
			TCHAR user[user_info_max_length_bytes/sizeof(TCHAR)];
			for (i = 0; i < inst->chats_num; i++) // try to find empty space
			{
				if (!inst->chats[i].hWnd) break;
			}
			if (i >= inst->chats_num) // allocate space
			{
				inst->chats_num += 20;
				chatwnd_t* ch = (chatwnd_t*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,inst->chats_num*sizeof(chatwnd_t));
				if (ch && inst->chats)
				{
					CopyMemory(ch, inst->chats, (inst->chats_num-20)*sizeof(chatwnd_t));
				}
				if (inst->chats) HeapFree(GetProcessHeap(),0,inst->chats);
				inst->chats = ch;
				i = inst->chats_num-20;
			}
			LeaveCriticalSection(&inst->chat_cs);
			memcpy(inst->chats[i].id,pac->nd->id,NETWORK_ID_LENGTH*sizeof(u32));
			inst->chats[i].hWnd = StartChat(inst->hMainWnd,pac->nd);
			inst->net->get_user_info(pac->nd,user);
			SetWindowText(inst->chats[i].hWnd,user);
			*/
		}
		else
			LeaveCriticalSection(&inst->chat_cs);
		SendMessage(inst->chats[i].hWnd,WM_PROCESSDATA,pac->buf_len-sizeof(u32),(LPARAM)&pac->buffer[1]);
        break;
	case NETWORK_PROT_FILE_TRANSFER:
		inst->ft->parse_packet(&pac->buffer[1], pac->buf_len - sizeof(u32), pac->nd);
		break;
	default:
	    // another protocols here
		break;
	}
}

//----------------------------------------------------------------------------
void RequestConnection(NODE* nd)
{
	u32 cmd = 0;
	PACKET_FOR_SEND pack;
	pack.buffer = &cmd;
	pack.buf_len = sizeof(u32);
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	getNet()->send_packet(&pack);
}
//----------------------------------------------------------------------------
bool isTrustUser(u32* id)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(id,ptr->nd.id)) return ptr->trust;
	}
	return false;
}
//----------------------------------------------------------------------------
void SetTrustUser(HWND hWnd, u32* id, bool trust)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(id, ptr->nd.id))
		{
			if (trust)
			{
				TCHAR id_str[100];
				TCHAR ip_str[20];
				TCHAR msg[1000];
				id2str(ptr->nd.id, id_str);
				ip2str(ptr->nd.ip, ip_str);
				_stprintf_s(msg, 1000, TEXT("Do you want to authorize following user:\nName: %s\nOriginal name: %s\nID: %s\nIP: %s\nPort: %d"), ptr->name, ptr->name_orig, id_str, ip_str, ptr->nd.port);
				if (IDYES != MessageBox(hWnd, msg, TEXT("Question"), MB_YESNO | MB_ICONQUESTION))
				{
					break;
				}
			}
			ptr->trust = trust;
			break;
		}
	}
}
//----------------------------------------------------------------------------
void appendLike(NODE* nd)
{
	TCHAR UserName[200];
	inst->net->get_user_info(nd, UserName);
	inst->usr->append_like(nd, UserName, UserName, false);
}
//----------------------------------------------------------------------------
void removeLike(u32* id)
{
	inst->usr->remove_like(id);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK RenameDialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	like_t* ptr;

	switch (uMsg)
	{
	case WM_INITDIALOG:
		ptr = (like_t*)lParam;
		SetProp(hWnd, TEXT("like"), ptr);
		if (ptr->name_orig) SetDlgItemText(hWnd, IDC_USER_ORIG_NAME, ptr->name_orig);
		if (ptr->name) SetDlgItemText(hWnd, IDC_USER_NAME, ptr->name);
		SetFocus(GetDlgItem(hWnd, IDC_USER_NAME));
		return TRUE;
	case WM_DESTROY:
		RemoveProp(hWnd, TEXT("like"));
		return 0;
	case WM_CLOSE:
		EndDialog(hWnd, 0);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			{
				int len = GetWindowTextLength(GetDlgItem(hWnd, IDC_USER_NAME));
				ptr = (like_t*)GetProp(hWnd, TEXT("like"));
				if (ptr->name) delete[] ptr->name;
				ptr->name = new tchar_t[len + 1];
				GetDlgItemText(hWnd, IDC_USER_NAME, ptr->name, len+1);
				EndDialog(hWnd, 0);
				return TRUE;
			}
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
void renameLike(HWND hWnd, u32* id)
{
	like_t* ptr;
	for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		if (isIdIdentical(id, ptr->nd.id))
		{
			DialogBoxParam(GetModuleHandle(NULL), TEXT("RenameDialog"), hWnd, RenameDialogProc, (LPARAM)ptr);
			break;
		}
	}
}
//----------------------------------------------------------------------------
static void updateTable(HWND hWnd)
{
	like_t* ptr;
	TCHAR str[100];
	HWND hUsers = GetDlgItem(hWnd,IDC_USERS);
    int cnt = 0;
	int cur_sel = ListView_GetSelectionMark(hUsers);
	int num = ListView_GetItemCount(hUsers);
	u32 sel_id[NETWORK_ID_LENGTH];
	if (!ListView_GetSelectedCount(hUsers))
	{
		cur_sel = -1;
	}
	if (cur_sel >= 0)
	{
		ListView_GetItemText(hUsers,cur_sel,3,str,100);
		str2id(str,sel_id);
	}
	for(ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
	{
		LVITEM item = {0};

		inst->net->update_ip(&ptr->nd);
			
		item.mask = LVIF_TEXT|LVIF_IMAGE;
		if (ptr->name) item.pszText = ptr->name;
		else  item.pszText = ptr->name_orig;
		if (isTrustUser(ptr->nd.id))
		{
			item.iImage = inst->net->is_user_online(ptr->nd.id) ? 0 : 1;
		}
		else
		{
			item.iImage = 2;
		}
		item.iItem = cnt++;
		if (num < cnt) item.iItem = ListView_InsertItem(hUsers,&item);
		else ListView_SetItem(hUsers,&item);
		item.mask = LVIF_TEXT;
		item.iSubItem = 1;
		item.pszText = str;
		ip2str(ptr->nd.ip,str);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 2;
		_stprintf_s(str,100,TEXT("%d"),ptr->nd.port);
		ListView_SetItem(hUsers,&item);
		item.iSubItem = 3;
		id2str(ptr->nd.id,str);
		ListView_SetItem(hUsers,&item);

		if ((cur_sel >= 0) && isIdIdentical(sel_id,ptr->nd.id))
		{
			ListView_SetItemState(hUsers,item.iItem,LVIS_SELECTED,LVIS_SELECTED);
			ListView_SetSelectionMark(hUsers,item.iItem);
		}
	}
	// clear all other items if exist
	while(cnt < (num = ListView_GetItemCount(hUsers)))
	{
		ListView_DeleteItem(hUsers,num-1);
	}
	ListView_RedrawItems(hUsers,0,cnt);
}
//----------------------------------------------------------------------------
VOID TrayMenu(HWND hWnd)
{
	HMENU menu = CreatePopupMenu();
	MENUITEMINFO item;
	POINT pt;
	UINT cmd;
	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_TYPE|MIIM_ID;
	item.fType = MFT_STRING;
	item.wID = IDM_ACTIVATE;
	item.dwTypeData = TEXT("ZAS Telephone");
	InsertMenuItem(menu, 0, TRUE, &item);
	item.wID = IDM_ABOUT;
	item.dwTypeData = TEXT("About");
	InsertMenuItem(menu,1,TRUE,&item);
	item.wID = IDM_EXIT;
	item.dwTypeData = TEXT("Exit");
	InsertMenuItem(menu, 2, TRUE, &item);
	GetCursorPos(&pt);
	cmd = TrackPopupMenu(menu,TPM_RETURNCMD,pt.x,pt.y,0,hWnd,NULL);
	switch (cmd)
	{
	case IDM_ACTIVATE:
		ShowWindow(hWnd, SW_NORMAL);
		SetForegroundWindow(hWnd);
		break;
	default:
		PostMessage(hWnd,WM_COMMAND,cmd,0);
		break;
	}
}
//----------------------------------------------------------------------------
static void userMenu(HWND hWnd,LPNMITEMACTIVATE ntf)
{
	TCHAR str[100];
	HMENU menu = CreatePopupMenu();
	MENUITEMINFO item;
	NODE nd;

	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),ntf->iItem,3,str,100);
	if (!str2id(str,nd.id)) return;
	ListView_GetItemText(GetDlgItem(hWnd, IDC_USERS), ntf->iItem, 1, str, 100);
	str2ip(str, &nd.ip);
	ListView_GetItemText(GetDlgItem(hWnd, IDC_USERS), ntf->iItem, 2, str, 100);
	nd.port = (u16)wcstol(str, NULL, 10);

	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_TYPE|MIIM_ID;
	item.fType = MFT_STRING;
	if (!isTrustUser(nd.id))
	{
		item.wID = IDM_ADDTRUST;
		item.dwTypeData = TEXT("Authorize");
		InsertMenuItem(menu, 0, TRUE, &item);
	}
	item.wID = IDM_REMOVESAVED;
	item.dwTypeData = TEXT("Remove from list");
	InsertMenuItem(menu,1,TRUE,&item);
	item.wID = IDM_REQUEST_CONNECTION;
	item.dwTypeData = TEXT("Request Connection");
	InsertMenuItem(menu, 2, TRUE, &item);
	item.wID = IDM_RENAME;
	item.dwTypeData = TEXT("Rename");
	InsertMenuItem(menu, 3, TRUE, &item);
	item.wID = IDM_TRANSFERFILE;
	item.dwTypeData = TEXT("Send file");
	InsertMenuItem(menu, 4, TRUE, &item);
	ClientToScreen(GetDlgItem(hWnd, IDC_USERS), &ntf->ptAction);
	switch (TrackPopupMenu(menu,TPM_RETURNCMD,ntf->ptAction.x,ntf->ptAction.y,0,hWnd,NULL))
	{
	case IDM_ADDTRUST:
		SetTrustUser(hWnd,nd.id,true);
		break;
	case IDM_REMOVESAVED:
		inst->usr->remove_like(nd.id);
		break;
	case IDM_RENAME:
		renameLike(hWnd,nd.id);
		break;
	case IDM_TRANSFERFILE:
		FileTransfer(hWnd,&nd);
		break;
	case IDM_REQUEST_CONNECTION:
		RequestConnection(&nd);
		break;
	}
}
//----------------------------------------------------------------------------
static void selectUser(HWND hWnd, u32* id)
{
	TCHAR str[100];
	u32 item_id[NETWORK_ID_LENGTH];
	for (int i = 0; i < ListView_GetItemCount(GetDlgItem(hWnd,IDC_USERS)); i++)
	{
		ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),i,3,str,100);
		if (!str2id(str,item_id)) return;
		if (isIdIdentical(id,item_id))
		{
			ListView_SetItemState(GetDlgItem(hWnd,IDC_USERS),i,LVIS_SELECTED,LVIS_SELECTED);
			ListView_SetSelectionMark(GetDlgItem(hWnd,IDC_USERS),i);
			break;
		}
	}
}
//----------------------------------------------------------------------------
void StartChat(u32* id)
{
	TCHAR UserName[200];
	u32 i;
	NODE nd;
	settings_t* set = (settings_t*)GetProp(inst->hMainWnd, TEXT("set"));
	if (!set->allowAnyone)
	{
		like_t* ptr;
		bool in_list = false;
		for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
		{
			if (isIdIdentical(id, ptr->nd.id))
			{
				in_list = true;
				break;
			}
		}
		if (!in_list) return;
	}
	// try to find same window
	EnterCriticalSection(&inst->chat_cs);
	for (i = 0; i < inst->chats_num; i++)
	{
		if (inst->chats[i].hWnd && isIdIdentical(id,inst->chats[i].id))
		{
			SetForegroundWindow(inst->chats[i].hWnd);
			LeaveCriticalSection(&inst->chat_cs);
			return;
		}
	}
	// try to find empty space
	for (i = 0; i < inst->chats_num; i++)
	{
		if (!inst->chats[i].hWnd) break;
	}
	if (i >= inst->chats_num)
	{
		// allocate space
		inst->chats_num += 20;
		chatwnd_t* ch = (chatwnd_t*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,inst->chats_num*sizeof(chatwnd_t));
		if (ch && inst->chats)
		{
			CopyMemory(ch, inst->chats, (inst->chats_num-20)*sizeof(chatwnd_t));
		}
		if (inst->chats) HeapFree(GetProcessHeap(),0,inst->chats);
		inst->chats = ch;
	}
	LeaveCriticalSection(&inst->chat_cs);
	if (inst->net->findNode(id,&nd) || inst->kad->findNode(id,&nd))
	{
		memcpy(inst->chats[i].id,nd.id,NETWORK_ID_LENGTH*sizeof(u32));
		inst->chats[i].hWnd = StartChat(inst->hMainWnd,&nd);
		inst->net->get_user_info(&nd,UserName);
		SetWindowText(inst->chats[i].hWnd,UserName);
	}
}
//----------------------------------------------------------------------------
static void ActivateChat(HWND hWnd, int iItem)
{
	TCHAR str[100];
	u32 id[NETWORK_ID_LENGTH];
	ListView_GetItemText(GetDlgItem(hWnd,IDC_USERS),iItem,3,str,100);
	if (str2id(str,id))
	{
		StartChat(id);
	}
}
//----------------------------------------------------------------------------
static void DeleteChat(HWND hWnd, HWND chat)
{
	u32 i;
	EnterCriticalSection(&inst->chat_cs);
	for (i = 0; i < inst->chats_num; i++)
	{
		if (inst->chats[i].hWnd == chat)
		{
			inst->chats[i].hWnd = NULL;
			break;
		}
	}
	LeaveCriticalSection(&inst->chat_cs);
}
//----------------------------------------------------------------------------
static void ShowAbout(HWND hWnd)
{
const LPTSTR about = TEXT("ZAS Telephone ") VERSION TEXT(" ") BUILD_NUM TEXT(" ") DATE_NOW TEXT("\r\n\r\n")
					TEXT("To all Freedom loving people of the World. ")
					TEXT("We do our best to provide you private communication in the upcoming darkness of the Orwellian age. ")
					TEXT("The software binaries are free and always be free for any use. The source code is included only for review purpose. ")
					TEXT("You are not allowed to modify any part of this software or source code or use it for your products unless we give you explicit permission.\r\n\r\n")
					TEXT("The software is provided as is. There are no warranties or liabilities of any kind.\r\n\r\n")
					TEXT("Report bugs and suggestions to:\r\n\r\n")
					TEXT("ZAS Telephone Team\r\n")
					TEXT("zas@zas-comm.ru\r\n\r\n")
					
TEXT("Key Type:  DSA/ELG\r\n")
TEXT("Algorithm: 3DES\r\n")
TEXT("\r\n")
TEXT("-----BEGIN PGP PUBLIC KEY BLOCK-----\r\n")
TEXT("Version: PGPfreeware 5.5.2 for non-commercial use <http://www.pgp.com>\r\n\r\n")
TEXT("mQGiBFLi8QwRBADN266rAA/91Lo9LZg8s7o5iR9xaM5Pi6dKLlGCHqsNoejuESRh\r\n")
TEXT("zTbussbnKDk2Py4FKpQBZJdLvWcJoFOjJw+Tz5zkAsn9dlSwebMw1r3GF3m1gU1w\r\n")
TEXT("jRRgGWMP09FG8IRiPo8ZC0cDVziH9CaTg7hrhaQWoSYfiCfSI0iISeN7MQCg/zdY\r\n")
TEXT("l7IqjjtG1iBO/ZmXtYui04ED/2nAAmU2IW5tY3FuAyKBgwVPrfqUl8+aSGI3Enk6\r\n")
TEXT("woIchGEd1Coq557Z8DGnUEHn8yoJVrMQCiYlro+BAqKOUn+Vm2kugYfcV8yDXLEl\r\n")
TEXT("rmW/VAksZ5VLZ+BCqf0pJI/MR9PINBuN7tQIHp7OmxQCx6TG2IhVYsrsXfvtY5iP\r\n")
TEXT("CAtwBACg2ZKPPpvUrllloXhFFEXQ7kLv6BzCGDoevEk7AtYM9eNrp590EtB4Bl6J\r\n")
TEXT("2A0TFh2MsEYDoCTc7EI1Cj5EoPj2biEBBDJ/DUQOHR58fHsz90LHZgo3fYmfc0Qu\r\n")
TEXT("kXQ1I5GCsBqrm7VNGF7e8jOt5hpMFGQMgkPkoUV/8Lj2iTdh1rQVWkFTIDx6YXNA\r\n")
TEXT("emFzLWNvbW0ucnU+iQBLBBARAgALBQJS4vEMBAsCAwEACgkQK4UXmS+niMG6SgCd\r\n")
TEXT("GYZnTxNlLXG20OHnfnDm/CwA8pEAoKhgsL4kgYgXqEA4SsEMLR9imjUQuQQNBFLi\r\n")
TEXT("8ecQEADrITWmV9/bhDr0UU1ssWbtRkV/5E1f28DM52ZjHOOtPROteSAiMcVG0AOq\r\n")
TEXT("1JAoOXf4+2qdou7foXkUYjmFdLZlk9oOSNXP57/Ll11YOyN7e917EJ7ZPd5/AXu9\r\n")
TEXT("TbF3KrPX74Fs5SX6dxuZxGkAxDKtlx08lAFKp8yWRpLD8DnKSfH0wx29L3tblpdI\r\n")
TEXT("x5u5onpwe1SbCMWVbCvwRArUrY+EVnCvvcTTJKlAs9Qrw3Ufn2VMHwCn2rVyHMy6\r\n")
TEXT("e6N3/denVqMzoXOjBYh1h8G9mMZimQ/mFmXAls6S8HsevmFYbXpkXv6g1+0z9Uap\r\n")
TEXT("TpoH5b65OJXRvYKupUwXkJe2KP/CLVGOREUiYisJofP/at/m5+kaPFcXw1ylz1kH\r\n")
TEXT("1HCDerQ+ka4TPILsiK3xhX3FQHHd3xIJHhE10B9mt5XVV9RcOKB5Cx6XYOSoQq3P\r\n")
TEXT("S8XJpcvHoQLa6j/4fVXgr20pGR+vx95bXD178UM7+debFpwW/lAis5RTt9sj0rFH\r\n")
TEXT("45W7SFkIgf70tPnYW8LiNpl8x+c394Eu9PZfcji2ymatD+VXga6mDrnioquAehKZ\r\n")
TEXT("C13TRKJMsbJd6/596bJQpATZ4EQ5VK7g43JFp/gMk8fywn8QsqcowocQZgPdfSZ7\r\n")
TEXT("L+wf0Z3j/aqqwINWRBw1FZoTReSltN2fMHk0xAQA9ZI6Cm13fwACAhAAn41qFqO6\r\n")
TEXT("lkfR91hQejaS5n5bq1+tAz+PcaQFnyD7+CQ/zvkhK+i2M3ahGP7J5MPkaBRbFdxT\r\n")
TEXT("PvOtaW9xliNP1vh8X5Fn1iNo+5KkohXzFOUevRF61lMR4KeBX46FAUH4LXxGJU7c\r\n")
TEXT("npJAbzKZvY+khX+Rwdmy28j12ryjaAU6s7cpHokjh2wRmy030yxrIrKNzHQWgSYX\r\n")
TEXT("5460QGoXQM9qps+Bxm0j9NHZyKrICAxJGY06JLUVy6FdMsUstT2bfTHpmekiOMQl\r\n")
TEXT("q889lXIpu9TOs+T9aKrCWi9T7TiV5M7oGqux6DNS8OqAihZVkaHA5HczyBqjVA9y\r\n")
TEXT("zA8svc1KvjTO+lFrofmH/cnfwZUazqlDXHOPi+psJ915LPxr1pSDAPCvB/qQT2td\r\n")
TEXT("o1dE6j+CrrmXJMoU/gmoFmLmDlFDNgvwCBDSpCFTYFpGd2rPHyqAUiBd5iiH9wZB\r\n")
TEXT("ge87nzrrdA0jTxop6WJ1TfF0OQW26vmMmeU6FcM3pk9GqRqDkam8HqjxIh5929v7\r\n")
TEXT("/41HqeYGn6zmGj/Oj0d0QKj6mgUSin+LotZxEGGPQU5ZgV3QVKyFoWaZ+Em5MX6u\r\n")
TEXT("kLwZyNr5O3CJuvJYvBbsfQaIF/YlL6V1H5ilDG+naqulatL7fXqQ/3r6F3IoDiAa\r\n")
TEXT("CppVKRxXNKloBfjMRryanTZXMR/8LW/wsISJAEUEGBECAAYFAlLi8ecACgkQK4UX\r\n")
TEXT("mS+niMGwWwCcC3rEZ5jNa1uPZD1iFZXGhPGpINIAmOV6u6q3zguMl+1wEdddLmHt\r\n")
TEXT("hEk=\r\n")
TEXT("=1JLW\r\n")
TEXT("-----END PGP PUBLIC KEY BLOCK-----\r\n");

	HIMAGELIST IL = ImageList_Create(128,128,ILC_COLOR32,1,1);
	ImageList_AddIcon(IL,LoadIcon(GetModuleHandle(NULL),MAKEINTRESOURCE(IDI_ABOUT)));
	SetProp(hWnd,TEXT("image"),IL);

	SetDlgItemText(hWnd,IDC_ABOUT_TEXT,about);

	SetFocus(GetDlgItem(hWnd, IDOK));
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK AboutDialog(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		ShowAbout(hWnd);
		return FALSE;
    case WM_DESTROY:
		ImageList_Destroy((HIMAGELIST)RemoveProp(hWnd,TEXT("image")));
		return 0;
    case WM_CLOSE:
		EndDialog(hWnd,0);
		return TRUE;

    case WM_PAINT:
	{
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(hWnd, &ps);
		ImageList_Draw((HIMAGELIST)GetProp(hWnd,TEXT("image")),0,hdc,2,40,ILD_NORMAL);
        EndPaint(hWnd, &ps);
		return FALSE;
	}
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			EndDialog(hWnd,0);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static void MainWndSizing(HWND hWnd, WPARAM wParam, LPRECT newWndRect)
{
	RECT oldWndRect;
	RECT oldClientRect;
	const LONG lMinWidth = 200;
	const LONG lMinHeight = 300;
	//
	GetWindowRect(hWnd, &oldWndRect);
	GetClientRect(hWnd, &oldClientRect);
	//
	oldClientRect.right -= oldWndRect.right - oldWndRect.left - (newWndRect->right - newWndRect->left);		// calculate new client width
	oldClientRect.bottom -= oldWndRect.bottom - oldWndRect.top - (newWndRect->bottom - newWndRect->top);	// calculate new client height
	// adjust width
	if (oldClientRect.right < lMinWidth){
		switch (wParam)
		{
		case WMSZ_BOTTOMLEFT:
		case WMSZ_LEFT:
		case WMSZ_TOPLEFT:
			newWndRect->left -= lMinWidth - oldClientRect.right;
			break;
			//
		default:
			newWndRect->right += lMinWidth - oldClientRect.right;
		}
	}
	// adjust height
	if (oldClientRect.bottom < lMinHeight){
		switch (wParam)
		{
		case WMSZ_TOPLEFT:
		case WMSZ_TOP:
		case WMSZ_TOPRIGHT:
			newWndRect->top -= lMinHeight - oldClientRect.bottom;
			break;
			//
		default:
			newWndRect->bottom += lMinHeight - oldClientRect.bottom;
		}
	}
}
//----------------------------------------------------------------------------
static void RequestConnectionInit(HWND hWnd, NODE* nd)
{
	TCHAR msg[1000];
	TCHAR id_str[100];
	TCHAR ip_str[20];
	TCHAR UserName[200];
	NODE* node = (NODE*)LocalAlloc(LPTR, sizeof(NODE));
	memcpy(node, nd, sizeof(NODE));
	SetProp(hWnd, TEXT("node"), node);
	id2str(nd->id, id_str);
	ip2str(nd->ip, ip_str);
	inst->net->get_user_info(nd, UserName);
	_stprintf_s(msg, 1000, TEXT("Do you accept connection request from user:\nName: %s\nID: %s\nIP: %s\nPort: %d"), UserName, id_str, ip_str, nd->port);
	SetDlgItemText(hWnd, IDC_USER_NAME, msg);
}
//----------------------------------------------------------------------------
void GetMyName(LPTSTR name)
{
	if (inst->sdata.data_is_initialized)
	{
		memcpy(name,inst->sdata.certificate.user_info,inst->sdata.certificate.user_info_length_bytes);
		name[inst->sdata.certificate.user_info_length_bytes/sizeof(TCHAR)] = 0;
	}
	else
	{
		name[0] = 0;
	}
}
//----------------------------------------------------------------------------
static void ShowMyInfo(HWND hWnd)
{
	TCHAR id_str[ID_LENGTH_CCH];
	TCHAR user_info[user_info_max_length_bytes/sizeof(TCHAR) + 1];
	TCHAR msg[2000];
	if (inst->sdata.data_is_initialized)
	{
		GetMyName(user_info);
		id2str(inst->sdata.certificate.fingerprint,id_str);
		_stprintf_s(msg,2000,TEXT("Your information:\nUser name: %s\nID: %s"),user_info,id_str);
	}
	else
	{
		cp_strcchcpy(msg,2000,_T("Key isn't loaded. Can't show information about you"));
	}
	MessageBox(hWnd,msg,TEXT("User Information"),MB_OK|MB_ICONINFORMATION);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK RequestConnection(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		RequestConnectionInit(hWnd, (NODE*)lParam);
		return TRUE;
	case WM_DESTROY:
		LocalFree(GetProp(hWnd, TEXT("node")));
		return 0;
	case WM_CLOSE:
		DestroyWindow(hWnd);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			appendLike((NODE*)GetProp(hWnd, TEXT("node")));
			DestroyWindow(hWnd);
			return TRUE;
		case IDCANCEL:
			DestroyWindow(hWnd);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------
static void MainWndSize(HWND hWnd)
{
	RECT rc;
	GetClientRect(hWnd,&rc);

	MoveWindow(GetDlgItem(hWnd,IDC_USERS),0,0,rc.right,rc.bottom,TRUE);
}
//----------------------------------------------------------------------------
static void MainWndInit(HWND hWnd)
{
	LVITEM item = {0};
	LVCOLUMN col;
	HWND hSplash = CreateDialog(GetModuleHandle(NULL),TEXT("About"),hWnd,AboutDialog);
//	HWND hUsers = CreateWindowEx(WS_EX_CLIENTEDGE,WC_LISTVIEW,TEXT(""),WS_CHILD|WS_VISIBLE|LVS_REPORT|LVS_SINGLESEL|LVS_SORTASCENDING,0,0,10,20,hWnd,(HMENU)IDC_USERS,GetModuleHandle(NULL),NULL);
	HWND hUsers = CreateWindowEx(WS_EX_CLIENTEDGE, WC_LISTVIEW, TEXT(""), WS_CHILD | WS_VISIBLE | LVS_LIST | LVS_SINGLESEL | LVS_SORTASCENDING, 0, 0, 10, 20, hWnd, (HMENU)IDC_USERS, GetModuleHandle(NULL), NULL);
	ListView_SetExtendedListViewStyle(hUsers, LVS_EX_FULLROWSELECT | LVS_EX_AUTOSIZECOLUMNS | LVS_EX_GRIDLINES);
	// create columns
	col.mask = LVCF_WIDTH|LVCF_TEXT;
	col.pszText = TEXT("Name");
	col.cx = 100;
	ListView_InsertColumn(hUsers,0,&col);
	col.pszText = TEXT("IP");
	ListView_InsertColumn(hUsers,1,&col);
	col.pszText = TEXT("Port");
	ListView_InsertColumn(hUsers,2,&col);
	col.pszText = TEXT("ID");
	col.cx = 500;
	ListView_InsertColumn(hUsers,3,&col);
	// add image list
	HIMAGELIST hLarge = ImageList_Create(GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), ILC_COLOR32, 3, 1);
	HIMAGELIST hSmall = ImageList_Create(GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), ILC_COLOR32, 3, 1);
	HICON hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_ONLINE));
    ImageList_AddIcon(hLarge, hiconItem);
    ImageList_AddIcon(hSmall, hiconItem);
    DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_OFFLINE));
	ImageList_AddIcon(hLarge, hiconItem);
    ImageList_AddIcon(hSmall, hiconItem);
    DestroyIcon(hiconItem);
	hiconItem = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_USER_UNAUTH));
	ImageList_AddIcon(hLarge, hiconItem);
	ImageList_AddIcon(hSmall, hiconItem);
	DestroyIcon(hiconItem);
	ListView_SetImageList(hUsers, hLarge, LVSIL_NORMAL);
	ListView_SetImageList(hUsers,hSmall,LVSIL_SMALL);
	// settings
	settings_t *set = (settings_t*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(settings_t));
	SetProp(hWnd,TEXT("set"),set);
	LoadSettings(hWnd,set);
	CreateDialogParam(GetModuleHandle(NULL),TEXT("Log"),hWnd,LogProc,(LPARAM)set);
	
	//instance
	inst = (instance_t*)VirtualAlloc(NULL,sizeof(instance_t),MEM_RESERVE|MEM_COMMIT,PAGE_READWRITE);
	if (!VirtualLock(inst,sizeof(instance_t))) // For security reason this memory must not be written to the pagefile
	{
		MessageBox(hWnd,TEXT("Can't lock secure data memory"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
	}
	inst->hMainWnd = hWnd;
	InitializeCriticalSection(&inst->cs);
	InitializeCriticalSection(&inst->chat_cs);
	LoadKey(hWnd,set->keyFileName,&inst->sdata);
	NODE me;
	CopyMemory(me.id,inst->sdata.certificate.fingerprint,NETWORK_ID_LENGTH*sizeof(u32));
	me.ip = MAKEIPADDRESS(127,0,0,1);
	me.port = set->localPort;
	inst->kad = new KADEMLIA(&me,hWnd);
	inst->net = new NETWORK(&me,hWnd);
	inst->usr = new USERS();
	inst->ft = new FILE_TRANSFER(hWnd);
	SetTimer(hWnd, 1, 1000, NULL);
	inst->eStopThreadPassword = CreateEvent(NULL,TRUE,FALSE,NULL);
	inst->thPassword = CreateThread(NULL,0,PasswordThread,inst,0,NULL);
	inst->eStop = CreateEvent(NULL,TRUE,FALSE,NULL);
	SetWaveParam(set);
	inst->restart = false;
	inst->thVoice = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)voiceProc,inst,0,NULL);
	if (set->enableUPnP)
	{
		if (!setupUPnP(set->localPort,set->remotePort,0))
		{
			MessageBox(hWnd,TEXT("Can't setup UPnP"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		}
	}
	inst->usr->load_nodes(inst->kad,set->userFileName);
	inst->usr->load_likes(set->likesFileName);
	CreateListsDialog(hWnd);
	DestroyWindow(hSplash);
	//ShowWindow(g_logWnd,SW_NORMAL);
	//ShowWindow(g_NetList,SW_NORMAL);
	//ShowWindow(g_KadList,SW_NORMAL);
}
//----------------------------------------------------------------------------
static void MainWndDestroy(HWND hWnd)
{
	settings_t *set = (settings_t*)RemoveProp(hWnd,TEXT("set"));
	DestroyWindow(g_NetList);
	DestroyWindow(g_KadList);
	KillTimer(hWnd,1);
	SetEvent(inst->eStopThreadPassword);
	SetEvent(inst->eStop);
	WaitForSingleObject(inst->thPassword,1000);
	WaitForSingleObject(inst->thVoice,1000);
	CloseHandle(inst->eStopThreadPassword);
	CloseHandle(inst->eStop);
	CloseHandle(inst->thPassword);
	CloseHandle(inst->thVoice);
	inst->usr->save_nodes(set->userFileName);
	inst->usr->save_likes(set->likesFileName);
	if (inst->chats)
	{
		u32 i;
		for (i = 0; i < inst->chats_num; i++)
		{
			if (inst->chats[i].hWnd) DestroyWindow(inst->chats[i].hWnd);
		}
		HeapFree(GetProcessHeap(),0,inst->chats);
	}
	delete inst->ft;
	delete inst->kad;
	delete inst->net;
	delete inst->usr;
	DeleteCriticalSection(&inst->cs);
	DeleteCriticalSection(&inst->chat_cs);
	SaveSettings(hWnd,set);
	ZeroMemory(inst->password,password_max_length); // clear password
	HeapFree(GetProcessHeap(),0,set);
	VirtualUnlock(inst,sizeof(instance_t));
	VirtualFree(inst,sizeof(instance_t),MEM_RELEASE);
	DestroyWindow(g_logWnd);
}
//----------------------------------------------------------------------------
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:
		MainWndInit(hWnd);
		return 0;
		//
    case WM_DESTROY:
		MainWndDestroy(hWnd);
		PostQuitMessage(0);
		return 0;
		//
	case WM_SIZE:
		MainWndSize(hWnd);
		return 0;
		//
	case WM_SIZING:
		MainWndSizing(hWnd, wParam, (LPRECT)lParam);
		return TRUE;
		//
	case WM_TIMER:
		switch (wParam)
		{
		case 1:
			updateTable(hWnd);
			return 0;
		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDM_EXIT:
			DestroyWindow(hWnd);
			return 0;
		case IDM_ABOUT:
			DialogBox(GetModuleHandle(NULL),TEXT("About"),hWnd,AboutDialog);
			return 0;
		case IDM_HELP:
			return 0;
		case IDM_SETTINGS:
			SettingDialog(hWnd,(settings_t*)GetProp(hWnd,TEXT("set")),inst->usr,inst->kad);
			return 0;
		case IDM_GENERATE:
			GenerateKey(hWnd,((settings_t*)GetProp(hWnd,TEXT("set")))->keyFileName);
			return 0;
		case IDM_BOOTSTRAP:
			if (inst->net)
			{
				inst->net->reset_sessions();
			}
			if (inst->kad)
			{
				inst->kad->start();
				LogMessage(module,TEXT("Bootstrap started"));
			}
			return 0;
		case IDM_SHOW_LOG:
			ShowWindow(g_logWnd,SW_SHOWNORMAL);
			SetForegroundWindow(g_logWnd);
			return 0;
		case IDM_SHOW_NET:
			ShowWindow(g_NetList,SW_SHOWNORMAL);
			SetForegroundWindow(g_NetList);
			return 0;
		case IDM_SHOW_KAD:
			ShowWindow(g_KadList,SW_SHOWNORMAL);
			SetForegroundWindow(g_KadList);
			return 0;
		case IDM_FIND:
			findUser(hWnd);
			return 0;
		case IDM_MY_INFO:
			ShowMyInfo(hWnd);
			return 0;
		}
		return TRUE;
	case WM_NOTIFY:
	{
		LPNMITEMACTIVATE ntf = (LPNMITEMACTIVATE)lParam;
		if (ntf->hdr.idFrom == IDC_USERS)
		{
			switch (ntf->hdr.code)
			{
			case NM_DBLCLK:
				ActivateChat(hWnd,ntf->iItem);
				break;
			case NM_RCLICK:
				userMenu(hWnd,ntf);
				break;
			}
		}
		return FALSE;
	}
	case WM_SHOWTEXT:
		return 0;
	case WM_CREATECHAT:
		StartChat((u32*)lParam);
		return 0;
	case WM_DELETECHAT:
		DeleteChat(hWnd,(HWND)lParam);
		return 0;
	case WM_USER_FOUND:
	{
		NODE nd;
		inst->kad->findNode((u32*)lParam,&nd);
		selectUser(hWnd,(u32*)lParam);
		delete[] (u32*)lParam;
		return TRUE;
	}
	case WM_FILETRANSFER:
	{
		settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
		transfer_t* tr = (transfer_t*)lParam;
		if (!set->acceptFT)
		{
			like_t* ptr;
			bool in_list = false;
			for (ptr = inst->usr->getLikes(); ptr; ptr = ptr->next)
			{
				if (isIdIdentical(tr->nd.id, ptr->nd.id))
				{
					in_list = true;
					break;
				}
			}
			if (!in_list)
			{
				file_transfer_cancel(tr);
				return TRUE;
			}
		}

		CreateDialogParam(GetModuleHandle(NULL), TEXT("FileTransfer"), hWnd, TransferProc, lParam);
		return TRUE;
	}
	case WM_REQUEST_CONNECTION:
	{
		settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
		NODE* nd = (NODE*)lParam;
		if (set->allowConnection)
		{
			CreateDialogParam(GetModuleHandle(NULL), TEXT("RequestConnection"), hWnd, RequestConnection, (LPARAM)nd);
		}
		return TRUE;
	}
	case WM_SYSTEMTRAY:
		switch (LOWORD(lParam))
		{
		case WM_LBUTTONDOWN:
			ShowWindow(hWnd, SW_NORMAL);
			SetForegroundWindow(hWnd);
			break;
		case WM_RBUTTONDOWN:
			TrayMenu(hWnd);
			break;
		}
		return TRUE;
	case WM_CLOSE:
		ShowWindow(hWnd, SW_HIDE);
		return TRUE;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}
//----------------------------------------------------------------------------
int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG AMsg;
	BOOL ret;
	WNDCLASSEX wc;
    INITCOMMONCONTROLSEX icex;
	HANDLE hMutex;
	NOTIFYICONDATA nid;

	// check that another copy of ZAS is running
	hMutex = CreateMutex(NULL, TRUE, TEXT("ZAS telephone"));
	if (!hMutex)
	{
		MessageBox(NULL,TEXT("Can't create mutex"),TEXT("Fatal Error"),MB_OK|MB_ICONERROR);
		return 0;
	}
	if (GetLastError() == ERROR_ALREADY_EXISTS)
	{
		MessageBox(NULL,TEXT("Another copy of ZAS telephone is running. This applicaton will be closed"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
		HWND wnd = FindWindow(TEXT("MainWindowClass"),TEXT("ZAS Telephone"));
		if (wnd)
		{
			ShowWindow(wnd,SW_NORMAL);
			SetForegroundWindow(wnd);
		}
		return 0;
	}

	// common controls
	icex.dwSize = sizeof(INITCOMMONCONTROLSEX);
   	icex.dwICC   = ICC_LISTVIEW_CLASSES|ICC_TAB_CLASSES|ICC_BAR_CLASSES|ICC_PROGRESS_CLASS|ICC_INTERNET_CLASSES;
   	InitCommonControlsEx(&icex);
	
	// window class
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN));
	wc.hbrBackground = (HBRUSH)(COLOR_GRAYTEXT + 1);
	wc.hInstance = hInst;
	wc.lpszMenuName = MAKEINTRESOURCE(ID_MENU);
	wc.lpszClassName = TEXT("MainWindowClass");
	wc.hIconSm = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN));

	if (!RegisterClassEx(&wc)) 
		return 0;

	HWND hwndMain = CreateWindow(TEXT("MainWindowClass"),TEXT("ZAS Telephone"),WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|WS_CLIPSIBLINGS,0,0,200,600,NULL,NULL,hInst,NULL);
	if (!hwndMain) return 0;

	ShowWindow(hwndMain,nCmdShow);

	PostMessage(hwndMain, WM_COMMAND, IDM_BOOTSTRAP, 0); // bootstrap immediately after start

	HACCEL hAccel = LoadAccelerators(hInst, MAKEINTRESOURCE(ID_ACCEL));

	ZeroMemory(&nid,sizeof(nid));
	nid.cbSize = sizeof(nid);
	nid.hWnd = hwndMain;
	nid.uFlags = NIF_ICON|NIF_TIP|NIF_MESSAGE;
	nid.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN));
	cp_strcpy_s(nid.szTip,TEXT("ZAS Telephone"));
	nid.uVersion = NOTIFYICON_VERSION;
	nid.uCallbackMessage = WM_SYSTEMTRAY;
	Shell_NotifyIcon(NIM_ADD,&nid);

	while ((ret = GetMessage(&AMsg, NULL, 0, 0)) != 0) {
		if (ret == -1)
			continue;
        if (!TranslateAccelerator(hwndMain, hAccel, &AMsg)) {
			BOOL trans = TRUE;
			u32 i;
			// check all modeless dialogs
			EnterCriticalSection(&inst->chat_cs);
			for (i = 0; (i < inst->chats_num) && trans; i++)
				trans = trans && !(inst->chats[i].hWnd && IsWindow(inst->chats[i].hWnd) && IsDialogMessage(inst->chats[i].hWnd,&AMsg));
			LeaveCriticalSection(&inst->chat_cs);
			if (trans && !IsDialogMessage(g_logWnd,&AMsg) && !IsDialogMessage(g_NetList,&AMsg) && !IsDialogMessage(g_KadList,&AMsg))
			{
				TranslateMessage(&AMsg);
				DispatchMessage(&AMsg);
			}
		}
	}

	Shell_NotifyIcon(NIM_DELETE,&nid);

	CloseHandle(hMutex);

	return 0;
}
