#ifndef NETWORK_H_INCLUDED
#define NETWORK_H_INCLUDED
//-----------------------------------------------------------------------
#include "types.h"
#include "diffie_helman.h"
#include "kademlia.h"
//-----------------------------------------------------------------------
static const u32 NETWORK_SIZE_UDP = 384;
static const u32 NETWORK_SIZE_SECURED = NETWORK_SIZE_UDP-NETWORK_ID_LENGTH*4-4;
static const u32 NETWORK_SIZE_UNSECURED = NETWORK_SIZE_SECURED-16;
static const u32 NETWORK_SIZE_USEFUL = NETWORK_SIZE_UNSECURED-4;
static const u32 NETWORK_SIZE_DH_MAX = 8192;
enum {
	NETWORK_PROT_KADEMLIA,
	NETWORK_PROT_CONTROL,
	NETWORK_PROT_FILE_TRANSFER
};
//-----------------------------------------------------------------------
#define SESSION_ENCODED		0x00
#define SESSION_REQUEST		0x01
#define SESSION_RESPONSE	0x02
#define SESSION_PING_REQ	0x03
#define SESSION_PING_RESP	0x04
#define SESSION_NAME_REQ	0x05
#define SESSION_NAME_RESP	0x06
//-----------------------------------------------------------------------
#define SESSION_LAUNCHED	0x10000000
#define SESSION_PINGED		0x20000000
#define SESSION_RESPONDED	0x40000000
#define SESSION_REQ_COUNT_MASK	0xFF
#define SESSION_REQ_COUNT_MAX	200
//-----------------------------------------------------------------------
typedef struct {
	u32 rec_flags;	// bit flags of received packets
	u32 buff[NETWORK_SIZE_DH_MAX/4];
	u32 last_buf[NETWORK_SIZE_DH_MAX/4];
	u32 sent_buf[NETWORK_SIZE_DH_MAX/4];
	u32 rec_size;
	u32 last_size;
	u32 sent_size;
	TX_SECURE_SESSION tx_session;
} ESTABLISHMENT;
//-----------------------------------------------------------------------
class CONNECTION {
public:
	u32 flags;
	NODE nd;
	u8 userInfo[user_info_max_length_bytes];
	u32 userInfoSize;
	u32 pack_num_est;  // number of packets for establishment
	u32 pack_num_sec;  // number of secured packets
	u32 pack_num_drop; // number of dropped packets
	u32 pack_num_sent; // number of sent packets
	u32 timeout;

	ESTABLISHMENT* est;
	INITIALIZED_SESSION ses;
	CONNECTION* next_connection;
	CONNECTION(const NODE* node);
	~CONNECTION();
	bool isEstablished(void) {return (flags&SESSION_LAUNCHED) == SESSION_LAUNCHED;};
};
//-----------------------------------------------------------------------
typedef struct {
	u32* buffer;
	u32 buf_len;
	NODE* nd;
} NET_PACKET;
//-----------------------------------------------------------------------
typedef struct {
	const NODE* nd;
	u32 protocol;
	const u32* buffer;
	u32 buf_len;
} PACKET_FOR_SEND;
//-----------------------------------------------------------------------
class NETWORK {
	u32 sending_buff[NETWORK_SIZE_DH_MAX/4];
	u32 sending_size;
	TX_SECURE_SESSION cmn_tx_session;
	CONNECTION* con;
	HWND mainWnd;
	HANDLE eStopThreads;
	HANDLE recThread;
	HANDLE tmoThread;
	SOCKET soc;
	CRITICAL_SECTION cs;
	u32 my_id[NETWORK_ID_LENGTH];
	sockaddr_in my_addr;

	friend DWORD WINAPI receiverProc(NETWORK* n);
	friend DWORD WINAPI timeoutProc(NETWORK* n);

	CONNECTION* find_connection(const NODE* nd, bool establish);
	void send_big_buffer(u32 ip, u16 port, u32* buf, u32 size, u32 command, u32* pack_cnt);
	bool collect_big_buffer(u32* cur_buf, u32 size, ESTABLISHMENT* est);

public:
	NETWORK(NODE* me, HWND mWnd);
	~NETWORK();

	u32 send_packet(PACKET_FOR_SEND* pack);

	CONNECTION* get_first_node(void) {return con;};
	CONNECTION* get_next_node(CONNECTION* cur);
	void get_user_info(NODE* nd, LPTSTR info);
	void get_counters(const NODE* nd, u32* est, u32* sec, u32* drop, u32* sent);
	bool is_user_online(const u32* id);
	void update_ip(NODE* nd);
	void reset_sessions(void);

	bool findUserByName(u8* userInfo, u16 size, u32* id);
	bool findNode(const u32 *id, NODE *nd);
};
//-----------------------------------------------------------------------
#endif //NETWORK_H_INCLUDED
